import sys
import os

cpu_index = 0
cpu_total = 1
job_index = 0

if len(sys.argv)>=3:
	cpu_index = int(sys.argv[1])
	cpu_total = int(sys.argv[2])
else:
	print("warning : arguments might be missing")

def add_job(func):
	global job_index
	if job_index == cpu_index:
		print(f"I'm running job {job_index}")
		func()
	job_index = (job_index+1)%cpu_total

###

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision

import numpy as np
import matplotlib.pyplot as plt
# plt.ion()

import h5py

torch.cuda.empty_cache()
π = np.pi
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

#################################

class CosLoss(nn.Module):
	def __init__(self): super().__init__();self.cossim = nn.CosineSimilarity()
	def forward(self,x,y): return 1. - self.cossim(x,y).mean()

class DotSim(nn.Module):
	def __init__(self):
		super(DotSim,self).__init__()
	def forward(self,x,y):
		# y is interpreted as a target, we normalize against y only :
		# y -> y/|y|
		# x -> x/|y|
		# thus we return ⟨x|y⟩/|y|²
		yns = (y*y).sum(dim=1).unsqueeze(1) # y_norm_squared
		return (x*y/yns).sum(dim=1)

class DotLoss(nn.Module):
	def __init__(self):
		super(DotLoss,self).__init__()
		self.dot=DotSim()
		self.mse=nn.MSELoss()
	def forward(self,x,y):
		return self.mse(self.dot(x,y),self.dot(y,y))

"""
	dim_i,dim_o    : dimension of the linear application we are modeling
	dim_sample     : how many dimension to sample per epoch
	epochs_rel     : epoch relative to Jacobian dimension : 1 epoch = sample at least dim_i*dim_o vectors
	                 an other interpretation : epochs_rel=64 means every dimension should be trained 64 times
	lr             : learning rate (for the jacobian)
	loss_mode      : "mix","cos","mse"
	opti_mode      : "adam","sgd"
	initial_messup : the jacobian is scaled at start (for cos mode)
"""
def simplest_analysis(
	dim_i=20,dim_o=30,dim_sample=0.1,
	epochs_rel=64,lr0=-2,lr1=-2,
	loss_mode="mix",opti_mode="adam",initial_messup=10.
	):
	N = int(np.ceil(dim_sample*dim_i*dim_o))
	epochs=int(epochs_rel*dim_i*dim_o/N)
	lr_ = 10**np.linspace(lr0,lr1,epochs)
	identifier = f"dim=[{dim_i},{dim_o}]-dim_sample={dim_sample}-initial_messup={initial_messup}-{loss_mode}-{opti_mode}"
	blackbox = nn.Linear(dim_i,dim_o).to(device)
	J        = nn.Linear(dim_o,dim_i).to(device)
	with torch.no_grad(): J.weight=nn.Parameter(initial_messup*J.weight) # scale it up on purpose
	optim = None
	if   opti_mode=="adam": optim = torch.optim.Adam(J.parameters(),lr=lr_[0],eps=1e-3)
	elif opti_mode=="sgd" : optim = torch.optim.SGD (J.parameters(),lr=lr_[0])
	else : print("error"); return
	plot_err_x = []
	hist = {
		"loss_mse": [],
		"loss_cos": [],
		"loss_dot": [],
		"err"     : [],
		"err_mean": [],
		"err_std" : [],
		"err_min" : [],
		"err_max" : [],
		"err_.05" : [],
		"err_.25" : [],
		"err_med" : [],
		"err_.75" : [],
		"err_.95" : [],
		"sgn_neg" : [],
	}
	hist_epoch_rel=[]
	mse_ = nn.MSELoss()
	cos_ = CosLoss()
	dot_ = DotLoss()
	for epoch in range(epochs):
		epoch_rel = epoch*N/(dim_i*dim_o)
		for g in optim.param_groups: g['lr'] = lr_[epoch]
		########## learn a little more
		J.train()
		ε = 1e-1                                   # amplitude of dx, for Ax has no effect
		x = torch.randn(N,dim_i).to(device)        # pick an evaluation point at random
		dx = (torch.randn(x.shape)*ε).to(device)   #   x+ε  -   x  =  ε
		df = blackbox(x+dx) - blackbox(x)          # A(x+ε) - A(x) = Aε
		grad_o = torch.randn(N,dim_o).to(device)   # pick a grad_o at random
		grad_i = J(grad_o)                         # J should become A.T
		should = (grad_i*dx).sum(dim=1).view(1,-1) # for subspace
		found  = (grad_o*df).sum(dim=1).view(1,-1) # for subspace (corresponding ones)
		loss_mse = mse_(should, found)
		loss_cos = cos_(should, found)
		loss_dot = dot_(should, found)
		# step
		optim.zero_grad()
		if   loss_mode=="mse": loss_mse.backward()
		elif loss_mode=="cos": loss_cos.backward()
		elif loss_mode=="dot": loss_dot.backward()
		elif loss_mode=="mse+cos": ((loss_mse + loss_cos)/2.).backward() # average
		elif loss_mode=="dot+cos": ((loss_dot + loss_cos)/2.).backward() # average
		else : print("error"); return
		optim.step()
		# loss history
		hist_epoch_rel.append(epoch_rel)
		hist["loss_mse"].append(loss_mse.item())
		hist["loss_cos"].append(loss_cos.item())
		hist["loss_dot"].append(loss_dot.item())
		# feedback
		error = torch.flatten(blackbox.weight.T-J.weight).cpu().detach().numpy()
		error_abs = np.array(sorted(np.abs(error)))
		error_abs_N = len(error_abs)
		error_sgn = np.sign((blackbox.weight.T*J.weight).cpu().detach().numpy())
		plot_err_x.append(epoch_rel*np.ones(error.shape)) # kind of stupid...
		hist["err"     ].append(error)
		hist["err_mean"].append(error_abs.mean())
		hist["err_std" ].append(error_abs.std())
		hist["err_min" ].append(error_abs[0])
		hist["err_max" ].append(error_abs[-1])
		hist["err_.05" ].append(error_abs[(error_abs_N* 5)//100])
		hist["err_.25" ].append(error_abs[(error_abs_N*25)//100])
		hist["err_med" ].append(error_abs[(error_abs_N*50)//100])
		hist["err_.75" ].append(error_abs[(error_abs_N*75)//100])
		hist["err_.95" ].append(error_abs[(error_abs_N*95)//100])
		hist["sgn_neg" ].append((error_sgn<0).sum()/dim_i/dim_o)
		########## save progress
		if (epoch == epochs-1) or (epoch%100 == 0): # only last epoch or at regular interval
			J.eval()
			os.makedirs("out",exist_ok=True)
			with h5py.File(f'out/{identifier}.h5', 'w') as file:
				# for k in hist: file.create_dataset(k, data=hist[k])
				# don't save everything, we can recompute
				for k in ["loss_mse","loss_cos","loss_dot","err"]: file.create_dataset(k, data=hist[k])
			def plog(x): return np.log(np.array(x))
			#
			plt.figure(figsize=(10,5))
			plt.tight_layout(pad=0)
			plt.plot(hist_epoch_rel,plog(hist["loss_mse"]),c="blue"  ,label="mse",linewidth=1)
			plt.plot(hist_epoch_rel,plog(hist["loss_cos"]),c=[0,.8,1],label="cos",linewidth=1)
			plt.plot(hist_epoch_rel,plog(hist["loss_dot"]),c=[0,.5,1],label="dot",linewidth=1)
			plt.xlabel("epoch relative")
			plt.ylabel("loss")
			plt.legend()
			plt.savefig(f'out/{identifier}-loss.png');plt.close()
			#
			plt.figure(figsize=(10,5))
			plt.scatter(plot_err_x,plog(np.abs(hist["err"])),c="red",label="error",alpha=1/255,s=1)
			plt.xlabel("epoch relative")
			plt.ylabel("error")
			plt.savefig(f'out/{identifier}-err-raw.png');plt.close()
			#
			plt.figure(figsize=(10,5))
			# plt.plot(plog(hist["err_min" ]),c="red",label="min",linewidth=.5)
			# plt.plot(plog(hist["err_max" ]),c="red",label="max",linewidth=.5)
			plt.plot(hist_epoch_rel,plog(hist["err_.05"]),c="red",label="5%",linewidth=.25)
			plt.plot(hist_epoch_rel,plog(hist["err_med"]),c="red",label="50% = median",linewidth=1)
			plt.plot(hist_epoch_rel,plog(hist["err_.95"]),c="red",label="95%",linewidth=.25)
			plt.fill_between(hist_epoch_rel, plog(hist["err_.25"]), plog(hist["err_.75"]), alpha=0.2, color='red')
			### μ,σ stuff
			plt.plot(hist_epoch_rel,plog(hist["err_mean"]),c="red",label="mean",linewidth=1,linestyle='dashed')
			# μ=np.array(hist["err_mean"])
			# σ=np.array(hist["err_std"])
			# plt.fill_between(range(len(μ)), plog(μ-σ), plog(μ+σ), alpha=0.2, color='red')
			plt.xlabel("epoch relative")
			plt.ylabel("error")
			plt.savefig(f'out/{identifier}-err.png');plt.close()
			#
			plt.figure(figsize=(10,5))
			plt.plot(hist_epoch_rel,hist["err_.05"],c="red",label="5%",linewidth=.25)
			plt.plot(hist_epoch_rel,hist["err_med"],c="red",label="50% = median",linewidth=1)
			plt.plot(hist_epoch_rel,hist["err_.95"],c="red",label="95%",linewidth=.25)
			plt.fill_between(hist_epoch_rel, hist["err_.25"], hist["err_.75"], alpha=0.2, color='red')
			### μ,σ stuff
			plt.plot(hist_epoch_rel,hist["err_mean"],c="red",label="mean",linewidth=1,linestyle='dashed')
			plt.xlabel("epoch relative")
			plt.ylabel("error")
			plt.savefig(f'out/{identifier}-err-linear.png');plt.close()
			#
			# plt.figure(figsize=(30,5))
			plt.figure(figsize=(10,2.5))
			plt.tight_layout(pad=0)
			plt.plot(hist_epoch_rel,hist["sgn_neg"]) # sould go to 0
			plt.xlabel("epoch relative")
			plt.ylabel("amount of wrong sign in J")
			plt.savefig(f'out/{identifier}-sgn.png');plt.close()

# cpu_index = 0
# cpu_total = 900000
# job_index = 0

for [dim_i,dim_o] in [[2,3],[20,30],[200,300]]:
	for initial_messup in [0.1,1.0,10]:
		for opti_mode in ["adam","sgd"]:
			for loss_mode in ["mse+cos","dot+cos","mse","cos","dot"]:
				for dim_sample in 2.**np.array([0,-2,-4,-6,-8]):
					add_job(lambda:
						simplest_analysis(
							dim_i = dim_i,
							dim_o = dim_o,
							dim_sample = dim_sample,
							epochs_rel = 1024,
							lr0 = -2,
							lr1 = -3,
							loss_mode = loss_mode,
							opti_mode = opti_mode,
							initial_messup = initial_messup,
						)
					)
